"""
![logo_small](https://user-images.githubusercontent.com/48654552/134991429-1109ad1d-92fa-423f-8ce1-3b777471cb3d.png)

### PBar module for displaying custom progress bars for Python 3.9+

[GitHub Repository](https://github.com/DarviL82/PBar)
"""

__version__ = "1.8.0"

from . bar import PBar, animate, taskWrapper, barHelper
from . sets import CharSet, FormatSet, ColorSet